package io.repro.flutter;

import android.app.Activity;
import android.app.Application;
import android.os.AsyncTask;
import android.os.Handler;
import android.os.Looper;

import androidx.annotation.NonNull;

import java.lang.ref.WeakReference;
import java.lang.reflect.Method;
import android.net.Uri;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import io.flutter.embedding.engine.plugins.FlutterPlugin;
import io.flutter.embedding.engine.plugins.activity.ActivityAware;
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding;
import io.flutter.plugin.common.MethodCall;
import io.flutter.plugin.common.MethodChannel;
import io.flutter.plugin.common.MethodChannel.MethodCallHandler;
import io.flutter.plugin.common.MethodChannel.Result;
import io.flutter.plugin.common.PluginRegistry.Registrar;

import io.repro.android.remoteconfig.RemoteConfigListener;
import io.repro.android.remoteconfig.RemoteConfigValue;
import io.repro.android.tracking.*;

import io.repro.android.Repro;
import io.repro.android.user.UserProfileGender;
import io.repro.android.newsfeed.NewsFeedEntry;
import io.repro.android.newsfeed.NewsFeedCampaignType;
import io.repro.android.BridgeUtils;

/** ReproFlutterPlugin */
@SuppressWarnings("unchecked")
public class ReproFlutterPlugin implements MethodCallHandler, FlutterPlugin, ActivityAware {
  private static WeakReference<Activity> sActivityRef = null;

  private static MethodChannel channel;

  private static ExecutorService sExecutor;
  private static Handler sHandler;

  private static Map<String, MethodChannelHandler> methodMap = new HashMap<>();
  public interface MethodChannelHandler {
    void call(Object arguments, Result result);
  }

  /** Plugin registration. */
  // for v1 Android embedding
  public static void registerWith(Registrar registrar) {
    channel = new MethodChannel(registrar.messenger(), "io.repro.flutter");
    channel.setMethodCallHandler(new ReproFlutterPlugin());
    sActivityRef = new WeakReference<>(registrar.activity());

    registerMethodMap();

    // Execute passRuntimeValues triggered by registerWith,
    // since this is the only initialization process that will be executed reliably in the plugin.
    new Handler(Looper.getMainLooper()).postDelayed(new Runnable() {
      @Override
      public void run() {
        channel.invokeMethod("passRuntimeValues_dart", null);
      }
    }, 1000); // Wait for initialization of MethodChannel on the Dart side.

    sExecutor = Executors.newSingleThreadExecutor();
    sHandler = new Handler();
  }

  // for v2 Android embedding
  @Override
  public void onAttachedToEngine(@NonNull FlutterPluginBinding binding) {
    channel = new MethodChannel(binding.getBinaryMessenger(), "io.repro.flutter");
    channel.setMethodCallHandler(new ReproFlutterPlugin());
    sExecutor = Executors.newSingleThreadExecutor();
    sHandler = new Handler();
    registerMethodMap();

    new Handler(Looper.getMainLooper()).postDelayed(new Runnable() {
      @Override
      public void run() {
        channel.invokeMethod("passRuntimeValues_dart", null);
      }
    }, 1000); // Wait for initialization of MethodChannel on the Dart side.
  }

  // for v2 Android embedding
  @Override
  public void onAttachedToActivity(@NonNull ActivityPluginBinding binding) {
    sActivityRef = new WeakReference<>(binding.getActivity());
  }

  // for v2 Android embedding
  @Override
  public void onDetachedFromEngine(@NonNull FlutterPluginBinding binding) {}

  // for v2 Android embedding
  @Override
  public void onDetachedFromActivityForConfigChanges() {
    sActivityRef.clear();
  }

  // for v2 Android embedding
  @Override
  public void onReattachedToActivityForConfigChanges(@NonNull ActivityPluginBinding binding) {
    sActivityRef = new WeakReference<>(binding.getActivity());
  }

  // for v2 Android embedding
  @Override
  public void onDetachedFromActivity() {
    sActivityRef.clear();
  }

  private static void registerMethodMap() {
    methodMap.put("getUserID", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        final String ret = getUserID();
        result.success(ret);
      }
    });
    methodMap.put("setUserID", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        setUserID(arguments);
        result.success(null);
      }
    });
    methodMap.put("getDeviceID", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        final String ret = getDeviceID();
        result.success(ret);
      }
    });
    methodMap.put("setLogLevel", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        setLogLevel(arguments);
        result.success(null);
      }
    });
    methodMap.put("optIn", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        optIn(arguments);
        result.success(null);
      }
    });
    methodMap.put("setStringUserProfile", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        setStringUserProfile(arguments);
        result.success(null);
      }
    });
    methodMap.put("setIntUserProfile", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        setIntUserProfile(arguments);
        result.success(null);
      }
    });
    methodMap.put("setDoubleUserProfile", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        setDoubleUserProfile(arguments);
        result.success(null);
      }
    });
    methodMap.put("setDateUserProfile", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        setDateUserProfile(arguments);
        result.success(null);
      }
    });
    methodMap.put("setUserGender", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        setUserGender(arguments);
        result.success(null);
      }
    });
    methodMap.put("setUserEmailAddress", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        setUserEmailAddress(arguments);
        result.success(null);
      }
    });
    methodMap.put("track", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        track(arguments);
        result.success(null);
      }
    });
    // event tracking
    methodMap.put("track", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        track(arguments);
        result.success(null);
      }
    });
    methodMap.put("trackViewContent", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        trackViewContent(arguments);
        result.success(null);
      }
    });
    methodMap.put("trackSearch", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        trackSearch(arguments);
        result.success(null);
      }
    });
    methodMap.put("trackAddToWishlist", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        trackAddToWishlist(arguments);
        result.success(null);
      }
    });
    methodMap.put("trackAddToCart", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        trackAddToCart(arguments);
        result.success(null);
      }
    });
    methodMap.put("trackInitiateCheckout", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        trackInitiateCheckout(arguments);
        result.success(null);
      }
    });
    methodMap.put("trackAddPaymentInfo", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        trackAddPaymentInfo(arguments);
        result.success(null);
      }
    });
    methodMap.put("trackPurchase", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        trackPurchase(arguments);
        result.success(null);
      }
    });
    methodMap.put("trackShare", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        trackShare(arguments);
        result.success(null);
      }
    });
    methodMap.put("trackCompleteRegistration", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        trackCompleteRegistration(arguments);
        result.success(null);
      }
    });
    methodMap.put("trackLead", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        trackLead(arguments);
        result.success(null);
      }
    });
    // NewsFeeds
    methodMap.put("getNewsFeeds", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        getNewsFeeds(arguments, result);
      }
    });
    methodMap.put("updateNewsFeeds", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        updateNewsFeeds(arguments, result);
      }
    });
    // enable/disableInAppMessage
    methodMap.put("enableInAppMessagesOnForegroundTransition", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        enableInAppMessagesOnForegroundTransition();
        result.success(null);
      }
    });
    methodMap.put("disableInAppMessagesOnForegroundTransition", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        disableInAppMessagesOnForegroundTransition();
        result.success(null);
      }
    });
    // Silver Egg
    methodMap.put("setSilverEggCookie", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        setSilverEggCookie(arguments);
        result.success(null);
      }
    });
    methodMap.put("setSilverEggProdKey", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        setSilverEggProdKey(arguments);
        result.success(null);
      }
    });
    methodMap.put("startListening_setOpenUrlCallback", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        startListening_setOpenUrlCallback(arguments);
        result.success(null);
      }
    });

    // UX Optimizer
    methodMap.put("startListening_remoteConfig_fetch", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        startListening_fetch(arguments);
        result.success(null);
      }
    });
    methodMap.put("cancelListening", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        cancelListening(arguments);
        result.success(null);
      }
    });
    methodMap.put("remoteConfig_get", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        Map<String, Object> remoteConfigValueMap = remoteConfig_get(arguments);
        result.success(remoteConfigValueMap);
      }
    });
    methodMap.put("remoteConfig_getAllValues", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        Map<String, Map<String, Object>> values = remoteConfig_getAllValues();
        result.success(values);
      }
    });
    methodMap.put("remoteConfig_getAllValuesWithPrefix", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        Map<String, Map<String, Object>> values = remoteConfig_getAllValuesWithPrefix(arguments);
        result.success(values);
      }
    });
    methodMap.put("remoteConfig_setDefaultsFromMap", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        remoteConfig_setDefaultsFromMap(arguments);
        result.success(null);
      }
    });
    methodMap.put("remoteConfig_setDefaultsFromJsonString", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        remoteConfig_setDefaultsFromJsonString(arguments);
        result.success(null);
      }
    });
    methodMap.put("remoteConfig_activateFetched", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        boolean res = remoteConfig_activateFetched();
        result.success(res);
      }
    });
    methodMap.put("remoteConfig_forceReset", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        remoteConfig_forceReset();
        result.success(null);
      }
    });
    methodMap.put("remoteConfig_getLocalDefaultValue", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        Map<String, Object> remoteConfigValue = remoteConfig_getLocalDefaultValue(arguments);
        result.success(remoteConfigValue);
      }
    });
    methodMap.put("passRuntimeValues_native", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        try {
          passRuntimeValues(arguments);
        } catch (Exception e) {
          result.error("ReproFlutterPlugin", "Failed to call passRuntimeValues", e);
        }
        result.success(null);
      }
    });
    methodMap.put("applicationShouldHandlePushNotification", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        applicationShouldHandlePushNotification(arguments, result);
      }
    });
    methodMap.put("isAlreadyReceivedPushNotification", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        isAlreadyReceivedPushNotification(arguments, result);
      }
    });
    methodMap.put("markPushNotificationReceived", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        markPushNotificationReceived(arguments, result);
      }
    });
    methodMap.put("setPushRegistrationID", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        setPushRegistrationID(arguments);
        result.success(null);
      }
    });
    methodMap.put("trackNotificationOpened", new MethodChannelHandler() {
      @Override
      public void call(Object arguments, Result result) {
        trackNotificationOpened(arguments);
        result.success(null);
      }
    });
  }

  // Callbacks
  private static Map<Integer, Runnable> callbackById = new HashMap<>();

  @Override
  public void onMethodCall(final MethodCall call, final Result result) {

    MethodChannelHandler method = methodMap.get(call.method);
    if (method == null) {
      result.notImplemented();
      return;
    }

    try {
      method.call(call.arguments, result);
    } catch (Exception e) {
      result.error("ReproFlutterPlugin", "An error occurred", e);
    }
  }

  // core/other

  private static String getUserID() {
    return Repro.getUserID();
  }

  private static void setUserID(final Object arguments) {
    Repro.setUserID((String) arguments);
  }

  private static String getDeviceID() {
    return Repro.getDeviceID();
  }

  private static void setLogLevel(final Object arguments) {
    final String logLevel = (String) arguments;
    if ("Debug".equals(logLevel)) {
      Repro.setLogLevel(android.util.Log.DEBUG);
    } else if ("Info".equals(logLevel)) {
      Repro.setLogLevel(android.util.Log.INFO);
    } else if ("Warn".equals(logLevel)) {
      Repro.setLogLevel(android.util.Log.WARN);
    } else if ("Error".equals(logLevel)) {
      Repro.setLogLevel(android.util.Log.ERROR);
    }
  }

  private static void enableInAppMessagesOnForegroundTransition() {
    final Activity activity = sActivityRef.get();
    // If the activity is null, the Main SDK will acquire and use the currently displayed activity.
    Repro.enableInAppMessagesOnForegroundTransition(activity);
  }

  private static void disableInAppMessagesOnForegroundTransition() {
    Repro.disableInAppMessagesOnForegroundTransition();
  }

  private static void optIn(final Object arguments) {
    Repro.optIn((Boolean) arguments);
  }

  // user profile

  private static void setStringUserProfile(final Object arguments) {
    final List list = (List) arguments;
    final String key = (String) list.get(0);
    final String value = (String) list.get(1);
    Repro.setStringUserProfile(key, value);
  }

  private static void setIntUserProfile(final Object arguments) {
    final List list = (List) arguments;
    final String key = (String) list.get(0);
    final Integer value = (Integer) list.get(1);
    Repro.setIntUserProfile(key, value);
  }

  private static void setDoubleUserProfile(final Object arguments) {
    final List list = (List) arguments;
    final String key = (String) list.get(0);
    final Double value = (Double) list.get(1);
    Repro.setDoubleUserProfile(key, value);
  }

  private static void setDateUserProfile(final Object arguments) {
    final List list = (List) arguments;
    final String key = (String) list.get(0);
    final Long value = (Long) list.get(1);
    Repro.setDateUserProfile(key, new Date(value));
  }

  private static void setUserGender(final Object arguments) {
    final String gender = (String) arguments;
    if ("Male".equals(gender)) {
      Repro.setUserGender(UserProfileGender.MALE);
    } else if ("Female".equals(gender)) {
      Repro.setUserGender(UserProfileGender.FEMALE);
    } else if ("Other".equals(gender)) {
      Repro.setUserGender(UserProfileGender.OTHER);
    }
  }

  private static void setUserEmailAddress(final Object arguments) {
    final String address = (String) arguments;
    Repro.setUserEmailAddress(address);
  }

  // event tracking

  private static void track(final Object argument) {
    final List list = (List) argument;
    final String name = (String) list.get(0);
    final Map map = (Map) list.get(1);

    Repro.track(name, map);
  }

  private static void trackViewContent(final Object argument) {
    final List list = (List) argument;
    final String contentId = (String) list.get(0);
    final Map map = (Map) list.get(1);
    final ViewContentProperties props = new ViewContentProperties();
    Object tmp;

    tmp = map.get("value");             if (tmp instanceof Double)  { props.setValue((Double) tmp); }
    tmp = map.get("currency");          if (tmp instanceof String)  { props.setCurrency((String) tmp); }
    tmp = map.get("content_name");      if (tmp instanceof String)  { props.setContentName((String) tmp); }
    tmp = map.get("content_category");  if (tmp instanceof String)  { props.setContentCategory((String) tmp); }
    tmp = map.get("extras");            if (tmp instanceof Map)     { props.setExtras((Map) tmp); }

    Repro.trackViewContent(contentId, props);
  }

  private static void trackSearch(final Object argument) {
    final Map map = (Map) argument;
    final SearchProperties props = new SearchProperties();
    Object tmp;

    tmp = map.get("content_id");        if (tmp instanceof String)  { props.setContentId((String) tmp); }
    tmp = map.get("value");             if (tmp instanceof Double)  { props.setValue((Double) tmp); }
    tmp = map.get("currency");          if (tmp instanceof String)  { props.setCurrency((String) tmp); }
    tmp = map.get("content_category");  if (tmp instanceof String)  { props.setContentCategory((String) tmp); }
    tmp = map.get("search_string");     if (tmp instanceof String)  { props.setSearchString((String) tmp); }
    tmp = map.get("extras");            if (tmp instanceof Map)     { props.setExtras((Map) tmp); }

    Repro.trackSearch(props);
  }

  private static void trackAddToWishlist(final Object argument) {
    final Map map = (Map) argument;
    final AddToWishlistProperties props = new AddToWishlistProperties();
    Object tmp;

    tmp = map.get("content_id");        if (tmp instanceof String)  { props.setContentId((String) tmp); }
    tmp = map.get("value");             if (tmp instanceof Double)  { props.setValue((Double) tmp); }
    tmp = map.get("currency");          if (tmp instanceof String)  { props.setCurrency((String) tmp); }
    tmp = map.get("content_name");      if (tmp instanceof String)  { props.setContentName((String) tmp); }
    tmp = map.get("content_category");  if (tmp instanceof String)  { props.setContentCategory((String) tmp); }
    tmp = map.get("extras");            if (tmp instanceof Map)     { props.setExtras((Map) tmp); }

    Repro.trackAddToWishlist(props);
  }

  private static void trackAddToCart(final Object argument) {
    final List list = (List) argument;
    final String contentId = (String) list.get(0);
    final Map map = (Map) list.get(1);
    final AddToCartProperties props = new AddToCartProperties();
    Object tmp;

    tmp = map.get("value");             if (tmp instanceof Double)  { props.setValue((Double) tmp); }
    tmp = map.get("currency");          if (tmp instanceof String)  { props.setCurrency((String) tmp); }
    tmp = map.get("content_name");      if (tmp instanceof String)  { props.setContentName((String) tmp); }
    tmp = map.get("content_category");  if (tmp instanceof String)  { props.setContentCategory((String) tmp); }
    tmp = map.get("extras");            if (tmp instanceof Map)     { props.setExtras((Map) tmp); }

    Repro.trackAddToCart(contentId, props);
  }

  private static void trackInitiateCheckout(final Object argument) {
    final Map map = (Map) argument;
    final InitiateCheckoutProperties props = new InitiateCheckoutProperties();
    Object tmp;

    tmp = map.get("content_id");        if (tmp instanceof String)  { props.setContentId((String) tmp); }
    tmp = map.get("value");             if (tmp instanceof Double)  { props.setValue((Double) tmp); }
    tmp = map.get("currency");          if (tmp instanceof String)  { props.setCurrency((String) tmp); }
    tmp = map.get("content_name");      if (tmp instanceof String)  { props.setContentName((String) tmp); }
    tmp = map.get("content_category");  if (tmp instanceof String)  { props.setContentCategory((String) tmp); }
    tmp = map.get("num_items");         if (tmp instanceof Integer) { props.setNumItems((Integer) tmp); }
    tmp = map.get("extras");            if (tmp instanceof Map)     { props.setExtras((Map) tmp); }

    Repro.trackInitiateCheckout(props);
  }

  private static void trackAddPaymentInfo(final Object argument) {
    final Map map = (Map) argument;
    final AddPaymentInfoProperties props = new AddPaymentInfoProperties();
    Object tmp;

    tmp = map.get("content_id");        if (tmp instanceof String)  { props.setContentId((String) tmp); }
    tmp = map.get("value");             if (tmp instanceof Double)  { props.setValue((Double) tmp); }
    tmp = map.get("currency");          if (tmp instanceof String)  { props.setCurrency((String) tmp); }
    tmp = map.get("content_category");  if (tmp instanceof String)  { props.setContentCategory((String) tmp); }
    tmp = map.get("extras");            if (tmp instanceof Map)     { props.setExtras((Map) tmp); }

    Repro.trackAddPaymentInfo(props);
  }

  private static void trackPurchase(final Object argument) {
    final List list = (List) argument;
    final String contentId = (String) list.get(0);
    final Double value = (Double) list.get(1);
    final String currency = (String) list.get(2);
    final Map map = (Map) list.get(3);
    final PurchaseProperties props = new PurchaseProperties();
    Object tmp;

    tmp = map.get("content_name");      if (tmp instanceof String)  { props.setContentName((String) tmp); }
    tmp = map.get("content_category");  if (tmp instanceof String)  { props.setContentCategory((String) tmp); }
    tmp = map.get("num_items");         if (tmp instanceof Integer) { props.setNumItems((Integer) tmp); }
    tmp = map.get("extras");            if (tmp instanceof Map)     { props.setExtras((Map) tmp); }

    Repro.trackPurchase(contentId, value, currency, props);
  }

  private static void trackShare(final Object argument) {
    final Map map = (Map) argument;
    final ShareProperties props = new ShareProperties();
    Object tmp;

    tmp = map.get("content_id");        if (tmp instanceof String)  { props.setContentId((String) tmp); }
    tmp = map.get("service_name");      if (tmp instanceof String)  { props.setServiceName((String) tmp); }
    tmp = map.get("content_name");      if (tmp instanceof String)  { props.setContentName((String) tmp); }
    tmp = map.get("content_category");  if (tmp instanceof String)  { props.setContentCategory((String) tmp); }
    tmp = map.get("extras");            if (tmp instanceof Map)     { props.setExtras((Map) tmp); }

    Repro.trackShare(props);
  }

  private static void trackCompleteRegistration(final Object argument) {
    final Map map = (Map) argument;
    final CompleteRegistrationProperties props = new CompleteRegistrationProperties();
    Object tmp;

    tmp = map.get("value");             if (tmp instanceof Double)  { props.setValue((Double) tmp); }
    tmp = map.get("currency");          if (tmp instanceof String)  { props.setCurrency((String) tmp); }
    tmp = map.get("content_name");      if (tmp instanceof String)  { props.setContentName((String) tmp); }
    tmp = map.get("status");            if (tmp instanceof String)  { props.setStatus((String) tmp); }
    tmp = map.get("extras");            if (tmp instanceof Map)     { props.setExtras((Map) tmp); }

    Repro.trackCompleteRegistration(props);
  }

  private static void trackLead(final Object argument) {
    final Map map = (Map) argument;
    final LeadProperties props = new LeadProperties();
    Object tmp;

    tmp = map.get("value");             if (tmp instanceof Double)  { props.setValue((Double) tmp); }
    tmp = map.get("currency");          if (tmp instanceof String)  { props.setCurrency((String) tmp); }
    tmp = map.get("content_name");      if (tmp instanceof String)  { props.setContentName((String) tmp); }
    tmp = map.get("content_category");  if (tmp instanceof String)  { props.setContentCategory((String) tmp); }
    tmp = map.get("extras");            if (tmp instanceof Map)     { props.setExtras((Map) tmp); }

    Repro.trackLead(props);
  }

  private static void setSilverEggCookie(final Object arguments) {
    Repro.setSilverEggCookie((String)arguments);
  }

  private static void setSilverEggProdKey(final Object arguments) {
    Repro.setSilverEggProdKey((String)arguments);
  }

  private static void updateNewsFeeds(final Object argument, final Result result) {
    final List jsonList = (List)argument;

    final Exception[] exception = new Exception[1];

    sExecutor.execute(new Runnable() {
      @Override
      public void run() {
        try {
          List<NewsFeedEntry> updateTargets = new ArrayList<>();
          for (Object jsonStr : jsonList) {
            NewsFeedEntry entry = new NewsFeedEntry((String)jsonStr);
            updateTargets.add(entry);
          }
          Repro.updateNewsFeeds(updateTargets);
        } catch (Exception e) {
          exception[0] = e;
        }

        sHandler.post(new Runnable() {
          @Override
          public void run() {
            if (exception[0] == null) {
              result.success(null);
            } else {
              result.error("Native Error", "An error occurred in 'updateNewsFeeds'", exception[0].getMessage());
            }
          }
        });
      }
    });
  }

  private static void getNewsFeeds(final Object argument, final Result result) {
    final List args = (List)argument;
    final int limit = (int)args.get(0);

    // args[1] is Integer or Long
    // see: https://api.flutter.dev/flutter/services/StandardMessageCodec-class.html
    Long offsetId = null;
    if (args.get(1) != null) {
      try {
        offsetId = (long)args.get(1);
      } catch (ClassCastException e) {
        offsetId = (long)(int)args.get(1);
      }
    }
    final Long finalOffsetId = offsetId;

    // nonnull, not empty
    final String strCampaignType = (String)args.get(2);
    final NewsFeedCampaignType campaignType = NewsFeedCampaignType.getType(strCampaignType);

    final Exception[] exception = new Exception[1];
    final List<Map<String, String>> newsFeedList = new ArrayList<>();

    sExecutor.execute(new Runnable() {
      @Override
      public void run() {

        List<NewsFeedEntry> newsFeeds;
        try {
          if (finalOffsetId == null) {
            newsFeeds = Repro.getNewsFeeds(limit, campaignType);
          }
          else {
            newsFeeds = Repro.getNewsFeeds(limit, (long)finalOffsetId, campaignType);
          }

          for (NewsFeedEntry entry : newsFeeds) {
            newsFeedList.add(NewsFeedUtils.convertToMap(entry));
          }
        } catch (Exception e) {
          exception[0] = e;
        }

        sHandler.post(new Runnable() {
          @Override
          public void run() {
            if (exception[0] == null) {
              result.success(newsFeedList);
            } else {
              result.error("Native Error", "An error occurred in 'getNewsFeeds'", exception[0].getMessage());
            }
          }
        });
      }
    });

  }


  private static void applicationShouldHandlePushNotification(final Object argument, final Result result) {
    if (!(argument instanceof Map)) {
      result.error("Native Error", "`applicationShouldHandlePushNotification` received an invalid value.", null);
      return;
    }
    Map<String, String> data = (Map<String, String>)argument;
    boolean shouldHandle = Repro.applicationShouldHandlePushNotification(BridgeUtils.getApplication(), data);
    result.success(shouldHandle);
  }

  private static void isAlreadyReceivedPushNotification(final Object argument, final Result result) {
    if (!(argument instanceof Map)) {
      result.error("Native Error", "`isAlreadyReceivedPushNotification` received an invalid value.", null);
      return;
    }
    Map<String, String> data = (Map<String, String>)argument;
    boolean isReceived = Repro.isAlreadyReceivedPushNotification(BridgeUtils.getApplication(), data);
    result.success(isReceived);
  }

  private static void markPushNotificationReceived(final Object argument, final Result result) {
    if (!(argument instanceof Map)) {
      result.error("Native Error", "`markPushNotificationReceived` received an invalid value.", null);
      return;
    }
    Map<String, String> data = (Map<String, String>)argument;
    Repro.markPushNotificationReceived(BridgeUtils.getApplication(), data);
    result.success(null);
  }

  private static void setPushRegistrationID(final Object argument) {
    String token = (String)argument;
    Repro.setPushRegistrationID(token);
  }

  private static void trackNotificationOpened(final Object argument) {
    String notificationId = (String)argument;
    Repro.trackNotificationOpened(notificationId);
  }


  private static void startListening_setOpenUrlCallback(final Object argument) {
    final List args = (List)argument;
    final int currentListenerId = (int)args.get(0);
    Repro.setOpenUrlCallback(new Repro.OpenUrlCallback() {
      @Override
      public void onOpened(Uri uri) {
        final Uri linkUri = uri;
        callbackById.put(currentListenerId, new Runnable() {
          @Override
          public void run() {
            if (callbackById.containsKey(currentListenerId)) {
              final Map<String, Object> args = new HashMap<>();
              args.put("id", currentListenerId);
              args.put("args", linkUri.toString());
              new Handler(Looper.getMainLooper()).postDelayed(new Runnable() {
                @Override
                public void run() {
                  channel.invokeMethod("callListener", args);
                }
              }, 1000);
            }
          }
        });
        callbackById.get(currentListenerId).run();
      }
    });
  }


  // UX Optimizer
  private static void startListening_fetch(Object argument) {
    final List args = (List)argument;
    final int currentListenerId = (int)args.get(0);
    final int timeout = (int)args.get(1);

    Repro.getRemoteConfig().fetch(timeout, new RemoteConfigListener() {
      @Override
      public void onCompletion(FetchStatus fetchStatus) {
        final FetchStatus stat = fetchStatus;
        callbackById.put(currentListenerId, new Runnable() {
          @Override
          public void run() {
            if (callbackById.containsKey(currentListenerId)) {
              Map<String, Object> args = new HashMap<>();
              args.put("id", currentListenerId);
              args.put("args", stat.ordinal());
      
              channel.invokeMethod("callListener", args);
            }
          }
        });
        callbackById.get(currentListenerId).run();
      }
    });
  }
  // Callback post-processing
  private static void cancelListening(Object args) {
    // Get callback id
    int currentListenerId = (int) args;
    // Remove callback
    callbackById.remove(currentListenerId);
  }

  private static Map<String, Object> remoteConfig_get(Object argument) {
    final String key = (String)argument;
    RemoteConfigValue value = Repro.getRemoteConfig().get(key);
    return convertRemoteConfigValueToMap(value);
  }

  private static Map<String, Map<String, Object>> remoteConfig_getAllValues() {
    Map<String, RemoteConfigValue> values = Repro.getRemoteConfig().getAllValues();
    Map<String, Map<String, Object>> transferableMap = new HashMap<>();

    for(Map.Entry<String, RemoteConfigValue> entry : values.entrySet()){
      transferableMap.put(entry.getKey(), convertRemoteConfigValueToMap(entry.getValue()));
    }
    return transferableMap;
  }

  private static Map<String, Map<String, Object>> remoteConfig_getAllValuesWithPrefix(Object argument) {
    String prefix = (String)argument;
    Map<String, RemoteConfigValue> values = Repro.getRemoteConfig().getAllValuesWithPrefix(prefix);
    Map<String, Map<String, Object>> transferableMap = new HashMap<>();

    for(Map.Entry<String, RemoteConfigValue> entry : values.entrySet()){
      transferableMap.put(entry.getKey(), convertRemoteConfigValueToMap(entry.getValue()));
    }
    return transferableMap;
  }

  private static void remoteConfig_setDefaultsFromMap(Object arguments) {
    Map<String, Object> arg = (Map<String, Object>)arguments;
    Repro.getRemoteConfig().setDefaultsFromMap(arg);
  }

  private static void remoteConfig_setDefaultsFromJsonString(Object jsonString) {
    String json = (String)jsonString;
    Repro.getRemoteConfig().setDefaultsFromJsonString(json);
  }

  private static boolean remoteConfig_activateFetched() {
    return Repro.getRemoteConfig().activateFetched();
  }

  private static Map<String, Object> remoteConfig_getLocalDefaultValue(Object arguments) {
    final String key = (String)arguments;
    RemoteConfigValue value = Repro.getRemoteConfig().getLocalDefaultValue(key);
    return convertRemoteConfigValueToMap(value);
  }

  private static void remoteConfig_forceReset() {
    Repro.getRemoteConfig().forceReset();
  }

  private static Map<String, Object> convertRemoteConfigValueToMap(RemoteConfigValue remoteConfigValue) {
    Map<String, Object> map = new HashMap<>();
    map.put("String", remoteConfigValue.asString());
    map.put("toString", remoteConfigValue.toString());

    // If UX Optimizer supports additional types, please add them here.
    // e.g. map.put("int", remoteConfigValue.asInt());
    return map;
  }

  private static void passRuntimeValues(final Object arguments) throws Exception {
    final Map<String, Object> values = (Map<String, Object>) arguments;
    Method method = Repro.class.getDeclaredMethod("_passRuntimeValues", Map.class);
    method.setAccessible(true);
    method.invoke(null, values);
  }
  
}
