#import "ReproFlutterPlugin.h"
#import "Repro/Repro.h"
#import "Repro/RPRNewsFeedEntry.h"

#include <dispatch/dispatch.h>

@interface Repro (NonPublicApi)
+ (void)_passRuntimeValues:(nonnull NSDictionary<NSString *, NSString *> *)values;
@end

@implementation ReproFlutterPlugin

static NSMutableDictionary* callbackById;
static FlutterMethodChannel* channel;


+ (void)registerWithRegistrar:(NSObject<FlutterPluginRegistrar>*)registrar {
    channel = [FlutterMethodChannel
               methodChannelWithName:@"io.repro.flutter"
               binaryMessenger:[registrar messenger]];
    ReproFlutterPlugin* instance = [[ReproFlutterPlugin alloc] init];
    [registrar addMethodCallDelegate:instance channel:channel];

    // Wait for initialization of MethodChannel on the dart side.
    dispatch_after(dispatch_time(DISPATCH_TIME_NOW, (int64_t)(1.0 * NSEC_PER_SEC)), dispatch_get_main_queue(), ^{
        [channel invokeMethod:@"passRuntimeValues_dart" arguments: @{}];
    });
}




// core/other

+ (id)repro_bridge_getUserID:(id)arguments {
    return [Repro getUserID];
}

+ (id)repro_bridge_setUserID:(id)arguments {
    [Repro setUserID:arguments];
    return nil;
}

+ (id)repro_bridge_getDeviceID:(id)arguments {
    return [Repro getDeviceID];
}

+ (id)repro_bridge_setLogLevel:(id)arguments {
    NSString *logLevel = (NSString *)arguments;
    if ([@"Debug" isEqualToString:logLevel]) {
        [Repro setLogLevel:RPRLogLevelDebug];
    } else if ([@"Info" isEqualToString:logLevel]) {
        [Repro setLogLevel:RPRLogLevelInfo];
    } else if ([@"Warn" isEqualToString:logLevel]) {
        [Repro setLogLevel:RPRLogLevelWarn];
    } else if ([@"Error" isEqualToString:logLevel]) {
        [Repro setLogLevel:RPRLogLevelError];
    }
    return nil;
}

+ (id)repro_bridge_disableInAppMessagesOnForegroundTransition:(id)arguments {
    [Repro disableInAppMessagesOnForegroundTransition];
    return nil;
}

+ (id)repro_bridge_enableInAppMessagesOnForegroundTransition:(id)arguments {
    [Repro enableInAppMessagesOnForegroundTransition];
    return nil;
}

+ (id)repro_bridge_optIn:(id)arguments {
    [Repro optIn:[((NSNumber *)arguments) boolValue]];
    return nil;
}




// user profile

+ (id)repro_bridge_setStringUserProfile:(id)arguments {
    NSArray *array = (NSArray *)arguments;
    NSString *key = (NSString *)array[0];
    NSString *value = (NSString *)array[1];
    [Repro setStringUserProfile:value forKey:key];
    return nil;
}

+ (id)repro_bridge_setIntUserProfile:(id)arguments {
    NSArray *array = (NSArray *)arguments;
    NSString *key = (NSString *)array[0];
    NSNumber *value = (NSNumber *)array[1];
    [Repro setIntUserProfile:value.integerValue forKey:key];
    return nil;
}

+ (id)repro_bridge_setDoubleUserProfile:(id)arguments {
    NSArray *array = (NSArray *)arguments;
    NSString *key = (NSString *)array[0];
    NSNumber *value = (NSNumber *)array[1];
    [Repro setDoubleUserProfile:value.doubleValue forKey:key];
    return nil;
}

+ (id)repro_bridge_setDateUserProfile:(id)arguments {
    NSArray *array = (NSArray *)arguments;
    NSString *key = (NSString *)array[0];
    NSNumber *value = (NSNumber *)array[1];
    NSDate* date = [NSDate dateWithTimeIntervalSince1970:(value.longValue) / 1000.0];
    [Repro setDateUserProfile:date forKey:key];
    return nil;
}

+ (id)repro_bridge_setUserGender:(id)arguments {
    NSString *gender = (NSString *)arguments;
    if ([@"Male" isEqualToString:gender]) {
        [Repro setUserGender:RPRUserProfileGenderMale];
    } else if ([@"Female" isEqualToString:gender]) {
        [Repro setUserGender:RPRUserProfileGenderFemale];
    } else if ([@"Other" isEqualToString:gender]) {
        [Repro setUserGender:RPRUserProfileGenderOther];
    }
    return nil;
}

+ (id)repro_bridge_setUserEmailAddress:(id)arguments {
    NSString *address = (NSString *)arguments;
    [Repro setUserEmailAddress:address];
    return nil;
}




// event tracking

+ (id)repro_bridge_track:(id)arguments {
    NSArray *array = (NSArray *)arguments;
    NSString *name = (NSString *)array[0];
    NSDictionary *dict = (NSDictionary *)array[1];
    [Repro track:name properties:dict];
    return nil;
}

+ (id)repro_bridge_trackViewContent:(id)arguments {
    NSArray *array = (NSArray *)arguments;
    NSString *contentId = (NSString *)array[0];
    NSDictionary *dict = (NSDictionary *)array[1];
    RPRViewContentProperties *props = [[RPRViewContentProperties alloc] init];
    NSObject *tmp;

    tmp = dict[@"value"];               if ([tmp isKindOfClass:NSNumber.class])     { [props setValue:((NSNumber *)tmp).doubleValue]; }
    tmp = dict[@"currency"];            if ([tmp isKindOfClass:NSString.class])     { [props setCurrency:(NSString *)tmp]; }
    tmp = dict[@"content_name"];        if ([tmp isKindOfClass:NSString.class])     { [props setContentName:(NSString *)tmp]; }
    tmp = dict[@"content_category"];    if ([tmp isKindOfClass:NSString.class])     { [props setContentCategory:(NSString *)tmp]; }
    tmp = dict[@"extras"];              if ([tmp isKindOfClass:NSDictionary.class]) { [props setExtras:(NSDictionary *)tmp]; }

    [Repro trackViewContent:contentId properties:props];
    return nil;
}

+ (id)repro_bridge_trackSearch:(id)arguments {
    NSDictionary *dict = (NSDictionary *)arguments;
    RPRSearchProperties *props = [[RPRSearchProperties alloc] init];
    NSObject *tmp;

    tmp = dict[@"content_id"];          if ([tmp isKindOfClass:NSString.class])     { [props setContentID:(NSString *)tmp]; }
    tmp = dict[@"value"];               if ([tmp isKindOfClass:NSNumber.class])     { [props setValue:((NSNumber *)tmp).doubleValue]; }
    tmp = dict[@"currency"];            if ([tmp isKindOfClass:NSString.class])     { [props setCurrency:(NSString *)tmp]; }
    tmp = dict[@"content_category"];    if ([tmp isKindOfClass:NSString.class])     { [props setContentCategory:(NSString *)tmp]; }
    tmp = dict[@"search_string"];       if ([tmp isKindOfClass:NSString.class])     { [props setSearchString:(NSString *)tmp]; }
    tmp = dict[@"extras"];              if ([tmp isKindOfClass:NSDictionary.class]) { [props setExtras:(NSDictionary *)tmp]; }

    [Repro trackSearch:props];
    return nil;
}

+ (id)repro_bridge_trackAddToWishlist:(id)arguments {
    NSDictionary *dict = (NSDictionary *)arguments;
    RPRAddToWishlistProperties *props = [[RPRAddToWishlistProperties alloc] init];
    NSObject *tmp;

    tmp = dict[@"content_id"];          if ([tmp isKindOfClass:NSString.class])     { [props setContentID:(NSString *)tmp]; }
    tmp = dict[@"value"];               if ([tmp isKindOfClass:NSNumber.class])     { [props setValue:((NSNumber *)tmp).doubleValue]; }
    tmp = dict[@"currency"];            if ([tmp isKindOfClass:NSString.class])     { [props setCurrency:(NSString *)tmp]; }
    tmp = dict[@"content_name"];        if ([tmp isKindOfClass:NSString.class])     { [props setContentName:(NSString *)tmp]; }
    tmp = dict[@"content_category"];    if ([tmp isKindOfClass:NSString.class])     { [props setContentCategory:(NSString *)tmp]; }
    tmp = dict[@"extras"];              if ([tmp isKindOfClass:NSDictionary.class]) { [props setExtras:(NSDictionary *)tmp]; }

    [Repro trackAddToWishlist:props];
    return nil;
}

+ (id)repro_bridge_trackAddToCart:(id)arguments {
    NSArray *array = (NSArray *)arguments;
    NSString *contentId = (NSString *)array[0];
    NSDictionary *dict = (NSDictionary *)array[1];
    RPRAddToCartProperties *props = [[RPRAddToCartProperties alloc] init];
    NSObject *tmp;

    tmp = dict[@"value"];               if ([tmp isKindOfClass:NSNumber.class])     { [props setValue:((NSNumber *)tmp).doubleValue]; }
    tmp = dict[@"currency"];            if ([tmp isKindOfClass:NSString.class])     { [props setCurrency:(NSString *)tmp]; }
    tmp = dict[@"content_name"];        if ([tmp isKindOfClass:NSString.class])     { [props setContentName:(NSString *)tmp]; }
    tmp = dict[@"content_category"];    if ([tmp isKindOfClass:NSString.class])     { [props setContentCategory:(NSString *)tmp]; }
    tmp = dict[@"extras"];              if ([tmp isKindOfClass:NSDictionary.class]) { [props setExtras:(NSDictionary *)tmp]; }

    [Repro trackAddToCart:contentId properties:props];
    return nil;
}

+ (id)repro_bridge_trackInitiateCheckout:(id)arguments {
    NSDictionary *dict = (NSDictionary *)arguments;
    RPRInitiateCheckoutProperties *props = [[RPRInitiateCheckoutProperties alloc] init];
    NSObject *tmp;

    tmp = dict[@"content_id"];          if ([tmp isKindOfClass:NSString.class])     { [props setContentID:(NSString *)tmp]; }
    tmp = dict[@"value"];               if ([tmp isKindOfClass:NSNumber.class])     { [props setValue:((NSNumber *)tmp).doubleValue]; }
    tmp = dict[@"currency"];            if ([tmp isKindOfClass:NSString.class])     { [props setCurrency:(NSString *)tmp]; }
    tmp = dict[@"content_name"];        if ([tmp isKindOfClass:NSString.class])     { [props setContentName:(NSString *)tmp]; }
    tmp = dict[@"content_category"];    if ([tmp isKindOfClass:NSString.class])     { [props setContentCategory:(NSString *)tmp]; }
    tmp = dict[@"num_items"];           if ([tmp isKindOfClass:NSNumber.class])     { [props setNumItems:((NSNumber *)tmp).integerValue]; }
    tmp = dict[@"extras"];              if ([tmp isKindOfClass:NSDictionary.class]) { [props setExtras:(NSDictionary *)tmp]; }

    [Repro trackInitiateCheckout:props];
    return nil;
}

+ (id)repro_bridge_trackAddPaymentInfo:(id)arguments {
    NSDictionary *dict = (NSDictionary *)arguments;
    RPRAddPaymentInfoProperties *props = [[RPRAddPaymentInfoProperties alloc] init];
    NSObject *tmp;

    tmp = dict[@"content_id"];          if ([tmp isKindOfClass:NSString.class])     { [props setContentID:(NSString *)tmp]; }
    tmp = dict[@"value"];               if ([tmp isKindOfClass:NSNumber.class])     { [props setValue:((NSNumber *)tmp).doubleValue]; }
    tmp = dict[@"currency"];            if ([tmp isKindOfClass:NSString.class])     { [props setCurrency:(NSString *)tmp]; }
    tmp = dict[@"content_category"];    if ([tmp isKindOfClass:NSString.class])     { [props setContentCategory:(NSString *)tmp]; }
    tmp = dict[@"extras"];              if ([tmp isKindOfClass:NSDictionary.class]) { [props setExtras:(NSDictionary *)tmp]; }

    [Repro trackAddPaymentInfo:props];
    return nil;
}

+ (id)repro_bridge_trackPurchase:(id)arguments {
    NSArray *array = (NSArray *)arguments;
    NSString *contentId = (NSString *)array[0];
    NSNumber *value = (NSNumber *)array[1];
    NSString *currency = (NSString *)array[2];
    NSDictionary *dict = (NSDictionary *)array[3];
    RPRPurchaseProperties *props = [[RPRPurchaseProperties alloc] init];
    NSObject *tmp;

    tmp = dict[@"content_name"];        if ([tmp isKindOfClass:NSString.class])     { [props setContentName:(NSString *)tmp]; }
    tmp = dict[@"content_category"];    if ([tmp isKindOfClass:NSString.class])     { [props setContentCategory:(NSString *)tmp]; }
    tmp = dict[@"num_items"];           if ([tmp isKindOfClass:NSNumber.class])     { [props setNumItems:((NSNumber *)tmp).integerValue]; }
    tmp = dict[@"extras"];              if ([tmp isKindOfClass:NSDictionary.class]) { [props setExtras:(NSDictionary *)tmp]; }

    [Repro trackPurchase:contentId value:value.doubleValue currency:currency properties:props];
    return nil;
}

+ (id)repro_bridge_trackShare:(id)arguments {
    NSDictionary *dict = (NSDictionary *)arguments;
    RPRShareProperties *props = [[RPRShareProperties alloc] init];
    NSObject *tmp;

    tmp = dict[@"content_id"];          if ([tmp isKindOfClass:NSString.class])     { [props setContentID:(NSString *)tmp]; }
    tmp = dict[@"service_name"];        if ([tmp isKindOfClass:NSString.class])     { [props setServiceName:(NSString *)tmp]; }
    tmp = dict[@"content_name"];        if ([tmp isKindOfClass:NSString.class])     { [props setContentName:(NSString *)tmp]; }
    tmp = dict[@"content_category"];    if ([tmp isKindOfClass:NSString.class])     { [props setContentCategory:(NSString *)tmp]; }
    tmp = dict[@"extras"];              if ([tmp isKindOfClass:NSDictionary.class]) { [props setExtras:(NSDictionary *)tmp]; }

    [Repro trackShare:props];
    return nil;
}

+ (id)repro_bridge_trackCompleteRegistration:(id)arguments {
    NSDictionary *dict = (NSDictionary *)arguments;
    RPRCompleteRegistrationProperties *props = [[RPRCompleteRegistrationProperties alloc] init];
    NSObject *tmp;

    tmp = dict[@"value"];               if ([tmp isKindOfClass:NSNumber.class])     { [props setValue:((NSNumber *)tmp).doubleValue]; }
    tmp = dict[@"currency"];            if ([tmp isKindOfClass:NSString.class])     { [props setCurrency:(NSString *)tmp]; }
    tmp = dict[@"content_name"];        if ([tmp isKindOfClass:NSString.class])     { [props setContentName:(NSString *)tmp]; }
    tmp = dict[@"status"];              if ([tmp isKindOfClass:NSString.class])     { [props setStatus:(NSString *)tmp]; }
    tmp = dict[@"extras"];              if ([tmp isKindOfClass:NSDictionary.class]) { [props setExtras:(NSDictionary *)tmp]; }

    [Repro trackCompleteRegistration:props];
    return nil;
}

+ (id)repro_bridge_trackLead:(id)arguments {
    NSDictionary *dict = (NSDictionary *)arguments;
    RPRLeadProperties *props = [[RPRLeadProperties alloc] init];
    NSObject *tmp;

    tmp = dict[@"value"];               if ([tmp isKindOfClass:NSNumber.class])     { [props setValue:((NSNumber *)tmp).doubleValue]; }
    tmp = dict[@"currency"];            if ([tmp isKindOfClass:NSString.class])     { [props setCurrency:(NSString *)tmp]; }
    tmp = dict[@"content_name"];        if ([tmp isKindOfClass:NSString.class])     { [props setContentName:(NSString *)tmp]; }
    tmp = dict[@"content_category"];    if ([tmp isKindOfClass:NSString.class])     { [props setContentCategory:(NSString *)tmp]; }
    tmp = dict[@"extras"];              if ([tmp isKindOfClass:NSDictionary.class]) { [props setExtras:(NSDictionary *)tmp]; }

    [Repro trackLead:props];
    return nil;
}



// external service

+ (id)repro_bridge_setSilverEggCookie:(id)arguments {
    NSString* cookie = (NSString*)arguments;
    [Repro setSilverEggCookie:cookie];
    return nil;
}

+ (id)repro_bridge_setSilverEggProdKey:(id)arguments {
    NSString* prodKey = (NSString*)arguments;
    [Repro setSilverEggProdKey:prodKey];
    return nil;
}




// news feeds

+ (RPRCampaignType)helper_getNewsFeedCampaignType:(nullable NSString *)rawCampaignType
{
    if ([rawCampaignType isKindOfClass:NSString.class]) {
        if ([rawCampaignType isEqualToString:@"push_notification"]) {
            return RPRCampaignTypePushNotification;
        }
        if ([rawCampaignType isEqualToString:@"in_app_message"]) {
            return RPRCampaignTypeInAppMessage;
        }
        if ([rawCampaignType isEqualToString:@"web_message"]) {
            return RPRCampaignTypeWebMessage;
        }
        if ([rawCampaignType isEqualToString:@"all"]) {
            return RPRCampaignTypeAll;
        }
    }
    return RPRCampaignTypeUnknown;
}

+ (NSString *)helper_convertCampaignTypeToString:(RPRCampaignType)campaignType
{
    if (campaignType == RPRCampaignTypePushNotification) {
        return @"push_notification";
    }
    else if (campaignType == RPRCampaignTypeInAppMessage) {
        return @"in_app_message";
    }
    else if (campaignType == RPRCampaignTypeWebMessage) {
        return @"web_message";
    }
    return @"unknown";
}

+ (NSDictionary*)helper_convertToDictionary:(RPRNewsFeedEntry*)newsFeedEntry {
    NSString *strId = @(newsFeedEntry.ID).stringValue;
    NSURL *linkUrl = newsFeedEntry.linkUrl;
    NSString* linkUrlString = @"";
    if (linkUrl != nil && [linkUrl absoluteString] != nil) {
        linkUrlString = [linkUrl absoluteString];
    }

    NSURL *imageUrl = newsFeedEntry.imageUrl;
    NSString *imageUrlString = @"";
    if (imageUrl != nil && [imageUrl absoluteString] != nil) {
        imageUrlString = [imageUrl absoluteString];
    }

    NSDateFormatter* formatter = NSDateFormatter.alloc.init;
    formatter.locale           = [NSLocale.alloc initWithLocaleIdentifier:@"en_US_POSIX"];
    formatter.calendar         = [NSCalendar.alloc initWithCalendarIdentifier:NSCalendarIdentifierGregorian];
    formatter.dateFormat       = @"yyyy-MM-dd'T'HH:mm:ssZZZ";
    NSString* timeStamp        = [formatter stringFromDate:newsFeedEntry.deliveredAt];

    NSString* shownString = newsFeedEntry.shown ? @"true" : @"false";
    NSString* readString = newsFeedEntry.read ? @"true" : @"false";

    NSDictionary* dic = @{
        @"id": strId,
        @"campaignType": [self helper_convertCampaignTypeToString:newsFeedEntry.campaignType],
        @"deviceID": newsFeedEntry.deviceID,
        @"title": newsFeedEntry.title,
        @"summary": newsFeedEntry.summary,
        @"body": newsFeedEntry.body,
        @"linkUrl": linkUrlString,
        @"imageUrl": imageUrlString,
        @"deliveredAt": timeStamp,
        @"shown": shownString,
        @"read": readString

    };
    return dic;
}


// Normally, this returns NSArray<NSDictionary *> *
+ (id)repro_bridge_getNewsFeeds:(id)arguments {
    NSArray *array = (NSArray *)arguments;
    NSNumber *numLimit = array[0];
    NSNumber *numOffset = array[1];
    NSString *strCampaignType = array[2];

    int64_t limit = [numLimit longLongValue];

    RPRCampaignType campaignType = [self helper_getNewsFeedCampaignType:strCampaignType];

    NSArray<RPRNewsFeedEntry *> * newsFeeds = nil;
    NSError *error = nil;

    if (numOffset && ![numOffset isEqual:[NSNull null]]) {
        int64_t offset = [numOffset longLongValue];
        newsFeeds = [Repro getNewsFeeds:limit offsetID:offset campaignType:campaignType error:&error];
    }
    else {
        newsFeeds = [Repro getNewsFeeds:limit campaignType:campaignType error:&error];
    }

    if (error) {
        // When newsFeeds is FlutterError, an exception will be raised in MethodChannel#invoke.
        return [FlutterError errorWithCode:@"Native Error" message:@"An error occurred in 'getNewsFeeds'" details:error.localizedDescription];
    }

    NSMutableArray<NSDictionary *> * arrayDic = [[NSMutableArray alloc] init];
    for (RPRNewsFeedEntry* entry in newsFeeds) {
        NSDictionary* dic = [self helper_convertToDictionary:entry];
        [arrayDic addObject: dic];
    }
    return arrayDic;
}

+ (id)repro_bridge_updateNewsFeeds:(id)arguments {
    NSArray<NSDictionary *> *stringMapList = (NSArray<NSDictionary *> *)arguments;

    NSMutableArray<RPRNewsFeedEntry*>* updateTargets = [[NSMutableArray alloc] init];
    for (NSDictionary* dic in stringMapList) {
        RPRNewsFeedEntry* entry = [[RPRNewsFeedEntry alloc] initWithDictionary:dic];
        [updateTargets addObject:entry];
    }

    NSError* error = nil;
    [Repro updateNewsFeeds:updateTargets error:&error];
    if (error) {
        return [FlutterError errorWithCode:@"Native Error" message:@"An error occurred in 'updateNewsFeeds'" details:error.localizedDescription];
    } else {
        return nil;
    }
}




// remote convig

+ (NSDictionary *)helper_convertRemoteConfigValueToDictionary:(RPRRemoteConfigValue *)remoteConfigValue {
    NSString *str = [remoteConfigValue stringValue];
    NSDictionary *dic = @{
        @"String": str,
        @"toString": str == nil ? @"" : str
    };
    return dic;
}

+ (id)repro_bridge_remoteConfig_get:(id)arguments {
    NSString *key = (NSString *)arguments;
    RPRRemoteConfig* remoteConfig = [Repro remoteConfig];
    RPRRemoteConfigValue* remoteConfigValue = [remoteConfig valueForKey:key];
    return [self helper_convertRemoteConfigValueToDictionary:remoteConfigValue];
}

+ (id)repro_bridge_remoteConfig_getAllValues:(id)arguments {
    NSDictionary<NSString*, RPRRemoteConfigValue*>* result = [Repro.remoteConfig allValues];

    NSDictionary *newDic = [NSMutableDictionary dictionary];
    for (NSString *key in [result keyEnumerator]) {
        NSDictionary *convertedValue = [self helper_convertRemoteConfigValueToDictionary:[result valueForKey:key]];
        [newDic setValue:convertedValue forKey:key];
    }
    return newDic;
}

+ (id)repro_bridge_remoteConfig_getAllValuesWithPrefix:(id)arguments {
    NSString *prefix = (NSString*)arguments;

    NSDictionary<NSString*, RPRRemoteConfigValue*>* result = [Repro.remoteConfig allValuesWithPrefix:prefix];

    NSDictionary* newDic = [NSMutableDictionary dictionary];
    for (NSString* key in [result keyEnumerator]) {
        NSDictionary *convertedValue = [self helper_convertRemoteConfigValueToDictionary:[result valueForKey:key]];
        [newDic setValue:convertedValue forKey:key];
    }
    return newDic;
}

+ (id)repro_bridge_remoteConfig_setDefaultsFromMap:(id)arguments {
    NSDictionary *defaultsDictionary = (NSDictionary *)arguments;
    [Repro.remoteConfig setDefaultsFromDictionary:defaultsDictionary];
    return nil;
}

+ (id)repro_bridge_remoteConfig_setDefaultsFromJsonString:(id)arguments {
    NSString *jsonString = (NSString*)arguments;
    [Repro.remoteConfig setDefaultsFromJsonString:jsonString];
    return nil;
}

+ (id)repro_bridge_remoteConfig_activateFetched:(id)arguments {
    return [NSNumber numberWithBool:[Repro.remoteConfig activateFetched]];
}

+ (id)repro_bridge_remoteConfig_forceReset:(id)arguments {
    [Repro.remoteConfig forceReset];
    return nil;
}

+ (id)repro_bridge_remoteConfig_getLocalDefaultValue:(id)arguments {
    NSString *key = (NSString *)arguments;
    RPRRemoteConfigValue *value = [Repro.remoteConfig localDefaultValueForKey:key];
    return [self helper_convertRemoteConfigValueToDictionary:value];
}

+ (id)repro_bridge_startListening_remoteConfig_fetch:(id)arguments {
    NSArray* array = (NSArray *)arguments;
    NSString* currentListenerId = [(NSNumber*)array[0] stringValue];
    NSNumber* timeoutMillisecs = (NSNumber*)array[1];
    double timeoutSecs = [timeoutMillisecs longLongValue] * 0.001;

    if (!callbackById) {
        callbackById = [NSMutableDictionary new];
    }

    RPRRemoteConfig* remoteConfig = [Repro remoteConfig];

    [remoteConfig fetchWithTimeout:timeoutSecs completionHandler:^(RPRRemoteConfigFetchStatus status) {
        NSNumber* stat = [NSNumber numberWithInteger:status];

        [callbackById setObject:^() {
            if ([callbackById valueForKey:currentListenerId] != nil) {
                [channel invokeMethod:@"callListener"
                            arguments:@{
                                @"id" : [NSNumber numberWithInt:[currentListenerId intValue]],
                                @"args" : stat
                            }
                ];
            }
        } forKey:currentListenerId];
        void (^cb)(void) = [callbackById valueForKey:currentListenerId];
        cb();
    }];

    return nil;
}

+ (id)repro_bridge_startListening_setOpenUrlCallback:(id)arguments {
    NSArray* array = (NSArray *)arguments;
    NSString* currentListenerId = [(NSNumber*)array[0] stringValue];

    if (!callbackById) {
        callbackById = [NSMutableDictionary new];
    }

    [Repro setOpenUrlCallback:^void(NSURL *url) {
        [callbackById setObject:^() {
            if ([callbackById valueForKey:currentListenerId] != nil) {
                [channel invokeMethod:@"callListener"
                    arguments:@{
                        @"id" : [NSNumber numberWithInt:[currentListenerId intValue]],
                        @"args" : [url absoluteString]
                    }
                ];
            }
        } forKey:currentListenerId];
        void (^cb)(void) = [callbackById valueForKey:currentListenerId];
        cb();
    }];

    return nil;

}


+ (id)repro_bridge_cancelListening:(id)arguments {
    NSString *listenerId = [(NSNumber *)arguments stringValue];
    if (callbackById == nil || ![callbackById.allKeys containsObject:listenerId]) {
        return nil; // no callback set
    }
    [callbackById removeObjectForKey:listenerId];
    return nil;
}


// internal

+ (id)repro_bridge_passRuntimeValues_native:(id)arguments {
    NSDictionary<NSString *, NSString *> *dic = (NSDictionary<NSString *, NSString *> *)arguments;
    [Repro _passRuntimeValues:dic];
    return nil;
}



// bridge method dispatch

- (void)handleMethodCall:(FlutterMethodCall *)call result:(FlutterResult)result {
    SEL sel = NSSelectorFromString([@"repro_bridge_" stringByAppendingString: [call.method stringByAppendingString:@":"]]);
    if (![ReproFlutterPlugin respondsToSelector:sel]) {
        NSString *msg = [@"Repro Bridge: Method not supported: " stringByAppendingString: call.method];
        result([FlutterError errorWithCode:@"Native Error" message:msg details:nil]);
        return;
    }
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Warc-performSelector-leaks"
    result([ReproFlutterPlugin performSelector:sel withObject:call.arguments]);
#pragma clang diagnostic pop
}

@end
