part of io.repro.flutter;

/// This class is the equivalent of a news feed entry.
class NewsFeedEntry {
  late int _id;
  late NewsFeedCampaignType _campaignType;
  late String _deviceID;
  late String _title;
  late String _summary;
  late String _body;
  String? _linkUrl;
  String? _imageUrl;
  DateTime? _deliveredAt;
  String? _rawDeliveredAt;
  late bool shown;
  late bool read;

  int get id => _id;
  NewsFeedCampaignType get campaignType => _campaignType;
  String get deviceID => _deviceID;
  String get title => _title;
  String get summary => _summary;
  String get body => _body;
  String? get linkUrl => _linkUrl;
  String? get imageUrl => _imageUrl;

  /// This property is a UTC date.
  DateTime? get deliveredAt => _deliveredAt;
  String? get rawDeliveredAt => _rawDeliveredAt;

  NewsFeedEntry(Map<String, String?> entry) {

    _id = int.parse(_getValueOrDefault(entry, 'id', defaultValue: '-1'));
    _campaignType = DartNewsFeedUtils.convertStringToCampaignType(_getValueOrDefault(entry, 'campaignType'));
    _deviceID = _getValueOrDefault(entry, 'deviceID');
    _title = _getValueOrDefault(entry, 'title');
    _summary = _getValueOrDefault(entry, 'summary');
    _body = _getValueOrDefault(entry, 'body');

    _linkUrl = entry['linkUrl']; // nullable
    _imageUrl = entry['imageUrl']; // nullable

    String? strDeliveredAt = entry['deliveredAt'];
    if (strDeliveredAt != null) {
      _deliveredAt = DateTime.parse(strDeliveredAt); // nullable
    }
    _rawDeliveredAt = entry['deliveredAt']; // nullable

    shown = entry['shown'] == 'true';
    read = entry['read'] == 'true';
  }

  /// Convert String? value to String
  String _getValueOrDefault(Map<String, String?> entry, String key, {String defaultValue = ""}) {
    String? strValue = entry[key];
    return strValue ?? defaultValue;
  }

}
